from __future__ import annotations

import datetime
import inspect
import logging
from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING, Any

from opentelemetry import trace
from opentelemetry.semconv._incubating.attributes.code_attributes import (
    CODE_FILEPATH,
    CODE_LINENO,
)
from opentelemetry.semconv._incubating.attributes.messaging_attributes import (
    MESSAGING_DESTINATION_NAME,
    MESSAGING_MESSAGE_ID,
    MESSAGING_OPERATION_NAME,
    MESSAGING_OPERATION_TYPE,
    MESSAGING_SYSTEM,
    MessagingOperationTypeValues,
)
from opentelemetry.semconv.attributes.error_attributes import ERROR_TYPE
from opentelemetry.trace import SpanKind, format_span_id, format_trace_id

from plain.models import IntegrityError
from plain.utils import timezone

from .registry import JobParameters, jobs_registry

if TYPE_CHECKING:
    from .models import JobProcess, JobRequest

logger = logging.getLogger(__name__)
tracer = trace.get_tracer("plain.jobs")


class JobType(ABCMeta):
    """
    Metaclass allows us to capture the original args/kwargs
    used to instantiate the job, so we can store them in the database
    when we schedule the job.
    """

    def __call__(self, *args: Any, **kwargs: Any) -> Job:
        instance = super().__call__(*args, **kwargs)
        instance._init_args = args
        instance._init_kwargs = kwargs
        return instance


class Job(metaclass=JobType):
    @abstractmethod
    def run(self) -> None:
        pass

    def run_in_worker(
        self,
        *,
        queue: str | None = None,
        delay: int | datetime.timedelta | datetime.datetime | None = None,
        priority: int | None = None,
        retries: int | None = None,
        retry_attempt: int = 0,
        unique_key: str | None = None,
    ) -> JobRequest | None:
        from .models import JobRequest

        job_class_name = jobs_registry.get_job_class_name(self.__class__)

        if queue is None:
            queue = self.get_queue()

        with tracer.start_as_current_span(
            f"run_in_worker {job_class_name}",
            kind=SpanKind.PRODUCER,
            attributes={
                MESSAGING_SYSTEM: "plain.jobs",
                MESSAGING_OPERATION_TYPE: MessagingOperationTypeValues.SEND.value,
                MESSAGING_OPERATION_NAME: "run_in_worker",
                MESSAGING_DESTINATION_NAME: queue,
            },
        ) as span:
            try:
                # Try to automatically annotate the source of the job
                caller = inspect.stack()[1]
                source = f"{caller.filename}:{caller.lineno}"
                span.set_attributes(
                    {
                        CODE_FILEPATH: caller.filename,
                        CODE_LINENO: caller.lineno,
                    }
                )
            except (IndexError, AttributeError):
                source = ""

            parameters = JobParameters.to_json(self._init_args, self._init_kwargs)

            if priority is None:
                priority = self.get_priority()

            if retries is None:
                retries = self.get_retries()

            if delay is None:
                start_at = None
            elif isinstance(delay, int):
                start_at = timezone.now() + datetime.timedelta(seconds=delay)
            elif isinstance(delay, datetime.timedelta):
                start_at = timezone.now() + delay
            elif isinstance(delay, datetime.datetime):
                start_at = delay
            else:
                raise ValueError(f"Invalid delay: {delay}")

            if unique_key is None:
                unique_key = self.get_unique_key()

            if unique_key:
                # Only need to look at in progress jobs
                # if we also have a unique key.
                # Otherwise it's up to the user to use _in_progress()
                if running := self._in_progress(unique_key):
                    span.set_attribute(ERROR_TYPE, "DuplicateJob")
                    return running

            # Is recording is not enough here... because we also record for summaries!

            # Capture current trace context
            current_span = trace.get_current_span()
            span_context = current_span.get_span_context()

            # Only include trace context if the span is being recorded (sampled)
            # This ensures jobs are only linked to traces that are actually being collected
            if current_span.is_recording() and span_context.is_valid:
                trace_id = f"0x{format_trace_id(span_context.trace_id)}"
                span_id = f"0x{format_span_id(span_context.span_id)}"
            else:
                trace_id = None
                span_id = None

            try:
                job_request = JobRequest(
                    job_class=job_class_name,
                    parameters=parameters,
                    start_at=start_at,
                    source=source,
                    queue=queue,
                    priority=priority,
                    retries=retries,
                    retry_attempt=retry_attempt,
                    unique_key=unique_key,
                    trace_id=trace_id,
                    span_id=span_id,
                )
                job_request.save(
                    clean_and_validate=False
                )  # So IntegrityError is raised on unique instead of potentially confusing ValidationError...

                span.set_attribute(
                    MESSAGING_MESSAGE_ID,
                    str(job_request.uuid),
                )

                # Add job UUID to current span for bidirectional linking
                span.set_attribute("job.uuid", str(job_request.uuid))
                span.set_status(trace.StatusCode.OK)

                return job_request
            except IntegrityError as e:
                span.set_attribute(ERROR_TYPE, "IntegrityError")
                span.set_status(trace.Status(trace.StatusCode.ERROR, "Duplicate job"))
                logger.warning("Job already in progress: %s", e)
                return None

    def _in_progress(self, unique_key: str) -> list[JobRequest | JobProcess]:
        """Get all JobRequests and JobProcess that are currently in progress, regardless of queue."""
        from .models import JobProcess, JobRequest

        job_class_name = jobs_registry.get_job_class_name(self.__class__)

        job_requests = JobRequest.query.filter(
            job_class=job_class_name,
            unique_key=unique_key,
        )

        jobs = JobProcess.query.filter(
            job_class=job_class_name,
            unique_key=unique_key,
        )

        return list(job_requests) + list(jobs)

    def get_unique_key(self) -> str:
        """
        A unique key to prevent duplicate jobs from being queued.
        Enabled by returning a non-empty string.

        Note that this is not a "once and only once" guarantee, but rather
        an "at least once" guarantee. Jobs should still be idempotent in case
        multiple instances are queued in a race condition.
        """
        return ""

    def get_queue(self) -> str:
        return "default"

    def get_priority(self) -> int:
        """
        Return the default priority for this job.

        Higher numbers run first: 10 > 5 > 0 > -5 > -10
        - Use positive numbers for high priority jobs
        - Use negative numbers for low priority jobs
        - Default is 0
        """
        return 0

    def get_retries(self) -> int:
        return 0

    def get_retry_delay(self, attempt: int) -> int:
        """
        Calculate a delay in seconds before the next retry attempt.

        On the first retry, attempt will be 1.
        """
        return 0



from __future__ import annotations

import datetime
import logging
import signal
from typing import Any

import click

from plain.cli import register_cli
from plain.runtime import settings
from plain.utils import timezone

from .models import JobProcess, JobRequest, JobResult
from .registry import jobs_registry
from .scheduling import load_schedule
from .workers import Worker

logger = logging.getLogger("plain.jobs")


@register_cli("jobs")
@click.group()
def cli() -> None:
    pass


@cli.command()
@click.option(
    "queues",
    "--queue",
    default=["default"],
    multiple=True,
    type=str,
    help="Queue to process",
)
@click.option(
    "--max-processes",
    "max_processes",
    default=None,
    type=int,
    envvar="PLAIN_JOBS_WORKER_MAX_PROCESSES",
)
@click.option(
    "--max-jobs-per-process",
    "max_jobs_per_process",
    default=None,
    type=int,
    envvar="PLAIN_JOBS_WORKER_MAX_JOBS_PER_PROCESS",
)
@click.option(
    "--max-pending-per-process",
    "max_pending_per_process",
    default=10,
    type=int,
    envvar="PLAIN_JOBS_WORKER_MAX_PENDING_PER_PROCESS",
)
@click.option(
    "--stats-every",
    "stats_every",
    default=60,
    type=int,
    envvar="PLAIN_JOBS_WORKER_STATS_EVERY",
)
@click.option(
    "--reload",
    is_flag=True,
    help="Watch files and auto-reload worker on changes",
)
def worker(
    queues: tuple[str, ...],
    max_processes: int | None,
    max_jobs_per_process: int | None,
    max_pending_per_process: int,
    stats_every: int,
    reload: bool,
) -> None:
    """Run the job worker."""
    jobs_schedule = load_schedule(settings.JOBS_SCHEDULE)

    if reload:
        from plain.internal.reloader import Reloader

        # Track whether we should continue restarting
        should_restart = {"value": True}
        current_worker = {"instance": None}

        def file_changed(filename: str) -> None:
            if current_worker["instance"]:
                current_worker["instance"].shutdown()

        def signal_shutdown(signalnum: int, _: Any) -> None:
            should_restart["value"] = False
            if current_worker["instance"]:
                current_worker["instance"].shutdown()

        # Allow the worker to be stopped gracefully on SIGTERM/SIGINT
        signal.signal(signal.SIGTERM, signal_shutdown)
        signal.signal(signal.SIGINT, signal_shutdown)

        # Start file watcher once, outside the loop
        reloader = Reloader(callback=file_changed, watch_html=False)
        reloader.start()

        while should_restart["value"]:
            worker = Worker(
                queues=list(queues),
                jobs_schedule=jobs_schedule,
                max_processes=max_processes,
                max_jobs_per_process=max_jobs_per_process,
                max_pending_per_process=max_pending_per_process,
                stats_every=stats_every,
            )
            current_worker["instance"] = worker

            # Start processing jobs (blocks until shutdown)
            worker.run()

    else:
        worker = Worker(
            queues=list(queues),
            jobs_schedule=jobs_schedule,
            max_processes=max_processes,
            max_jobs_per_process=max_jobs_per_process,
            max_pending_per_process=max_pending_per_process,
            stats_every=stats_every,
        )

        def _shutdown(signalnum: int, _: Any) -> None:
            logger.info("Job worker shutdown signal received signalnum=%s", signalnum)
            worker.shutdown()

        # Allow the worker to be stopped gracefully on SIGTERM
        signal.signal(signal.SIGTERM, _shutdown)
        signal.signal(signal.SIGINT, _shutdown)

        # Start processing jobs
        worker.run()


@cli.command()
def clear() -> None:
    """Clear all completed job results in all queues."""
    cutoff = timezone.now() - datetime.timedelta(
        seconds=settings.JOBS_RESULTS_RETENTION
    )
    click.echo(f"Clearing job results created before {cutoff}")
    results = JobResult.query.filter(created_at__lt=cutoff).delete()
    click.echo(f"Deleted {results[0]} jobs")


@cli.command()
def stats() -> None:
    """Stats across all queues."""
    pending = JobRequest.query.count()
    processing = JobProcess.query.count()

    successful = JobResult.query.successful().count()
    errored = JobResult.query.errored().count()
    lost = JobResult.query.lost().count()

    click.secho(f"Pending: {pending}", bold=True)
    click.secho(f"Processing: {processing}", bold=True)
    click.secho(f"Successful: {successful}", bold=True, fg="green")
    click.secho(f"Errored: {errored}", bold=True, fg="red")
    click.secho(f"Lost: {lost}", bold=True, fg="yellow")


@cli.command()
def purge() -> None:
    """Delete all running and pending jobs regardless of queue."""
    if not click.confirm(
        "Are you sure you want to clear all running and pending jobs? This will delete all current Jobs and JobRequests"
    ):
        return

    deleted = JobRequest.query.all().delete()[0]
    click.echo(f"Deleted {deleted} job requests")

    deleted = JobProcess.query.all().delete()[0]
    click.echo(f"Deleted {deleted} jobs")


@cli.command()
@click.argument("job_class_name", type=str)
def run(job_class_name: str) -> None:
    """Run a job class directly (and not using a worker)."""
    job = jobs_registry.load_job(job_class_name, {"args": [], "kwargs": {}})
    click.secho("Loaded job: ", bold=True, nl=False)
    print(job)
    job.run()


@cli.command("list")
def list_jobs() -> None:
    """List all registered jobs."""
    for name, job_class in jobs_registry.jobs.items():
        click.secho(f"{name}", bold=True, nl=False)
        # Get description from class docstring
        description = job_class.__doc__.strip() if job_class.__doc__ else ""
        if description:
            click.secho(f": {description}", dim=True)
        else:
            click.echo("")



from __future__ import annotations

from datetime import timedelta
from typing import Any

from plain import models
from plain.admin.cards import Card
from plain.admin.views import (
    AdminModelDetailView,
    AdminModelListView,
    AdminViewset,
    register_viewset,
)
from plain.http import ResponseRedirect
from plain.runtime import settings

from .models import JobProcess, JobRequest, JobResult


def _td_format(td_object: timedelta) -> str:
    seconds = int(td_object.total_seconds())
    periods = [
        ("year", 60 * 60 * 24 * 365),
        ("month", 60 * 60 * 24 * 30),
        ("day", 60 * 60 * 24),
        ("hour", 60 * 60),
        ("minute", 60),
        ("second", 1),
    ]

    strings = []
    for period_name, period_seconds in periods:
        if seconds > period_seconds:
            period_value, seconds = divmod(seconds, period_seconds)
            has_s = "s" if period_value > 1 else ""
            strings.append(f"{period_value} {period_name}{has_s}")

    return ", ".join(strings)


class SuccessfulJobsCard(Card):
    title = "Successful"
    text = "View"

    def get_number(self) -> int:
        return JobResult.query.successful().count()

    def get_link(self) -> str:
        return JobResultViewset.ListView.get_view_url() + "?display=Successful"


class ErroredJobsCard(Card):
    title = "Errored"
    text = "View"

    def get_number(self) -> int:
        return JobResult.query.errored().count()

    def get_link(self) -> str:
        return JobResultViewset.ListView.get_view_url() + "?display=Errored"


class LostJobsCard(Card):
    title = "Lost"
    text = "View"  # TODO make not required - just an icon?

    def get_description(self) -> str:
        delta = timedelta(seconds=settings.JOBS_TIMEOUT)
        return f"Jobs are considered lost after {_td_format(delta)}"

    def get_number(self) -> int:
        return JobResult.query.lost().count()

    def get_link(self) -> str:
        return JobResultViewset.ListView.get_view_url() + "?display=Lost"


class RetriedJobsCard(Card):
    title = "Retried"
    text = "View"  # TODO make not required - just an icon?

    def get_number(self) -> int:
        return JobResult.query.retried().count()

    def get_link(self) -> str:
        return JobResultViewset.ListView.get_view_url() + "?display=Retried"


class WaitingJobsCard(Card):
    title = "Waiting"

    def get_number(self) -> int:
        return JobProcess.query.waiting().count()


class RunningJobsCard(Card):
    title = "Running"

    def get_number(self) -> int:
        return JobProcess.query.running().count()


@register_viewset
class JobRequestViewset(AdminViewset):
    class ListView(AdminModelListView):
        nav_section = "Jobs"
        nav_icon = "gear"
        model = JobRequest
        title = "Requests"
        fields = ["id", "job_class", "priority", "created_at", "start_at", "unique_key"]
        actions = ["Delete"]
        queryset_order = ["priority", "-start_at", "-created_at"]

        def perform_action(self, action: str, target_ids: list[int]) -> None:
            if action == "Delete":
                JobRequest.query.filter(id__in=target_ids).delete()

    class DetailView(AdminModelDetailView):
        model = JobRequest
        title = "Request"


@register_viewset
class JobProcessViewset(AdminViewset):
    class ListView(AdminModelListView):
        nav_section = "Jobs"
        nav_icon = "gear"
        model = JobProcess
        title = "Processes"
        fields = [
            "id",
            "job_class",
            "priority",
            "created_at",
            "started_at",
            "unique_key",
        ]
        actions = ["Delete"]
        cards = [
            WaitingJobsCard,
            RunningJobsCard,
        ]

        def perform_action(self, action: str, target_ids: list[int]) -> None:
            if action == "Delete":
                JobProcess.query.filter(id__in=target_ids).delete()

    class DetailView(AdminModelDetailView):
        model = JobProcess
        title = "Process"


@register_viewset
class JobResultViewset(AdminViewset):
    class ListView(AdminModelListView):
        nav_section = "Jobs"
        nav_icon = "gear"
        model = JobResult
        title = "Results"
        fields = [
            "id",
            "job_class",
            "priority",
            "created_at",
            "status",
            "retried",
            "is_retry",
        ]
        search_fields = [
            "uuid",
            "job_process_uuid",
            "job_request_uuid",
            "job_class",
        ]
        cards = [
            SuccessfulJobsCard,
            ErroredJobsCard,
            LostJobsCard,
            RetriedJobsCard,
        ]
        filters = [
            "Successful",
            "Errored",
            "Cancelled",
            "Lost",
            "Retried",
        ]
        actions = [
            "Retry",
        ]
        allow_global_search = False

        def get_initial_queryset(self) -> Any:
            queryset = super().get_initial_queryset()
            queryset = queryset.annotate(
                retried=models.Case(
                    models.When(retry_job_request_uuid__isnull=False, then=True),
                    default=False,
                    output_field=models.BooleanField(),
                ),
                is_retry=models.Case(
                    models.When(retry_attempt__gt=0, then=True),
                    default=False,
                    output_field=models.BooleanField(),
                ),
            )
            if self.display == "Successful":
                return queryset.successful()
            if self.display == "Errored":
                return queryset.errored()
            if self.display == "Cancelled":
                return queryset.cancelled()
            if self.display == "Lost":
                return queryset.lost()
            if self.display == "Retried":
                return queryset.retried()
            return queryset

        def get_fields(self) -> list[str]:
            fields = super().get_fields()
            if self.display == "Retried":
                fields.append("retries")
                fields.append("retry_attempt")
            return fields

        def perform_action(self, action: str, target_ids: list[int]) -> None:
            if action == "Retry":
                for result in JobResult.query.filter(id__in=target_ids):
                    result.retry_job(delay=0)
            else:
                raise ValueError("Invalid action")

    class DetailView(AdminModelDetailView):
        model = JobResult
        title = "Result"

        def post(self) -> ResponseRedirect:
            self.object.retry_job(delay=0)
            return ResponseRedirect(".")



from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING

from plain.logs import app_logger

if TYPE_CHECKING:
    from .models import JobProcess, JobResult


class AppLoggerMiddleware:
    def __init__(self, run_job: Callable[[JobProcess], JobResult]) -> None:
        self.run_job = run_job

    def __call__(self, job: JobProcess) -> JobResult:
        with app_logger.include_context(
            job_request_uuid=str(job.job_request_uuid), job_process_uuid=str(job.uuid)
        ):
            return self.run_job(job)
